/*
 * Copyright 2011 Chris de Vreeze
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package eu.cdevreeze.springtx

import scala.util.control.Exception._
import org.springframework.transaction.TransactionStatus
import org.springframework.transaction.support.{ TransactionCallback, TransactionCallbackWithoutResult }

/**
 * Implicit conversions that are useful in the context of a Spring
 * <code>TransactionTemplate</code>, to make the application code slightly more Scala-friendly.
 *
 * An implicit conversion makes it possible to invoke method toCallback on a function taking a
 * TransactionStatus, thus turning that function into a TransactionCallback. Analogously method
 * toCallbackWithoutResult turns a Unit returning function taking a TransactionStatus into a
 * TransactionCallbackWithoutResult.
 *
 * Moreover, this object contains rollback support that can be used in a <code>TransactionTemplate</code>.
 * This rollback support is based on the <code>scala.util.control.Exception</code> API.
 *
 * Note how this approach of combining Scala with the Spring TransactionTemplate is a
 * different approach than using the <code>@Transactional</code> annotation, replacing
 * "AOP proxying magic" by "functional programming".
 *
 * @author Chris de Vreeze
 */
object TransactionTemplateUtils {

  /** "Implicit" class containing the toCallback method for functions taking a TransactionStatus. */
  final class ToTransactionCallback[T](f: (TransactionStatus) => T) {
    def toCallback: TransactionCallback[T] =
      new TransactionCallback[T] {
        def doInTransaction(status: TransactionStatus): T = f(status)
      }
  }

  /** "Implicit" class containing the toCallbackWithoutResult method for functions taking a TransactionStatus. */
  final class ToTransactionCallbackWithoutResult(f: (TransactionStatus) => Unit) {
    def toCallbackWithoutResult: TransactionCallbackWithoutResult =
      new TransactionCallbackWithoutResult {
        def doInTransactionWithoutResult(status: TransactionStatus): Unit = f(status)
      }
  }

  /**
   * Adds a toCallback method that implicitly converts a <code>(TransactionStatus) => T</code>
   * function to a <code>TransactionCallback</code>.
   */
  implicit def toTransactionCallbackConverter[T](f: (TransactionStatus) => T): ToTransactionCallback[T] =
    new ToTransactionCallback[T](f)

  /**
   * Adds a toCallbackWithoutResult method that implicitly converts a <code>(TransactionStatus) => T</code>
   * function to a <code>TransactionCallbackWithoutResult</code>.
   */
  implicit def toTransactionCallbackWithoutResultConverter[T](f: (TransactionStatus) => T): ToTransactionCallbackWithoutResult =
    new ToTransactionCallbackWithoutResult({ (status: TransactionStatus) => f(status); () })

  /**
   * Creates a <code>Catch[T]</code> that invokes <code>status.setRollbackOnly()</code> before re-throwing the exception.
   */
  def rollbackOn[T](exceptions: Class[_]*)(status: TransactionStatus): Catch[T] = {
    handling[T](exceptions: _*) by { (t: Throwable) =>
      status.setRollbackOnly()
      throw t
    }
  }
}
